/*==============================================================================
Project:	POV-Ray

Version:	3

File:	TextEditor.h

Description:
	This file contains typedefs, function prototypes and extern var declarations
	that need to be exported from texteditor.c.
------------------------------------------------------------------------------
Authors:
	Jim Nitchals, Eduard [esp] Schwan
------------------------------------------------------------------------------
	from Persistence of Vision(tm) Ray Tracer
	Copyright 1996 Persistence of Vision Team
------------------------------------------------------------------------------
	NOTICE: This source code file is provided so that users may experiment
	with enhancements to POV-Ray and to port the software to platforms other 
	than those supported by the POV-Ray Team.  There are strict rules under
	which you are permitted to use this file.  The rules are in the file
	named POVLEGAL.DOC which should be distributed with this file. If 
	POVLEGAL.DOC is not available or for more info please contact the POV-Ray
	Team Coordinator by leaving a message in CompuServe's Graphics Developer's
	Forum.  The latest version of POV-Ray may be found there as well.

	This program is based on the popular DKB raytracer version 2.12.
	DKBTrace was originally written by David K. Buck.
	DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
------------------------------------------------------------------------------
Change History:
	921110	[esp]	Created
	931001	[esp]	version 2.0 finished (Released on 10/4/93)
==============================================================================*/

#if !defined (TEXTEDITOR_H)
#define TEXTEDITOR_H


/*==== POV headers ====*/
#include "config.h"


/*==== Mac toolbox headers ====*/

#include <Controls.h>	// ControlHandle
#include <Dialogs.h>	// DialogPtr


/*==== External globals ====*/

#include "TE32K.h"

extern WindowPtr		gSrcWind_Window;
extern Boolean			gSrcWind_dirty;
extern Boolean			gSrcWind_visible;
extern Str255 			gSrcWind_FileName;
extern short			gSrcWind_VRefNum;
extern TE32KHandle		gSrcWind_TE32KH;


/*==== TextEditor.c Prototypes ====*/

int		SetUpFiles(void);
int		DoFile(int item);
int		SaveAs(Str255 fn, short *vRef);
int		SaveFile(Str255 fn, short vRef);
int		NewFile(Str255 fn, short *vRef);
int		OldFile(Str255 fn, short *vRef);
int		CreateFile(Str255 fn, short *vRef, short *theRef);
int		WriteFile(short refNum, char *p, long num);
int		ReadFile(short refNum);
int		pStrCopy(StringPtr p1, StringPtr p2);
int		FileError(Str255 s, Str255 f);
void	PreInitWindows(void);
int		SetUpWindows(void);
int		SetVScroll(void);
int		ShowSelect(void);
void	SelectAllText(void);
int		SetView(WindowPtr w);
int		UpdateWindow(WindowPtr theWindow);
void	ScrollEditHome(void);
void	ScrollEditEnd(void);
void	ScrollEditPage(short thePartCode);
int		DoContent(WindowPtr theWindow, EventRecord *theEvent);

long	GetSrcWindselStart(void);
long	GetSrcWindselEnd(void);
long	GetSrcWindteLength(void);
Handle	GetSrcWindhText(void);
void	SrcWindCutSelToClipboard(void);
void	SrcWindPasteFromClipboard(void);
void	SrcWindCopySelToClipboard(void);
void	SrcWindClearSel(void);
void	SrcWindSetSelect(long selStart, long selEnd);
void	SrcWindDeleteAllText(void);
void	SrcWindTEKey(char theKey);
void	SrcWindTEDeleteLeftCharKey(void);
void	SrcWindTEDeleteRightCharKey(void);
void	DoSrcWindTextInsert(char * someTextPtr, long someTextSize);
void	GotoSrcWindLine(short line_to_go);
void	MatchBracesAtCurSel(void);

void	MyResizeWindow(WindowPtr w, short h, short v);
int		CloseMyWindow(void);
int		DoActivateEditor(Boolean becomingActive);
void	Init_Editor(void); 
int		DoEditMouseDown(int windowPart, WindowPtr whichWindow, EventRecord *myEvent);
void	SrcWindMaintainCursor(void);

void SrcWindUndo(void);
void SrcWindRedo(void);


// ==============================================
// Check if undo is available
#define CheckSrcWindCanUndo() Check_CanUndo()

// ==============================================
// Check if redo is available
#define CheckSrcWindCanRedo() Check_CanRedo()


#endif // TEXTEDITOR_H
